#include <linux/types.h>
#include <linux/ip.h>
#include <linux/timer.h>
#include <linux/netfilter.h>
#include <linux/netfilter_ipv4.h>
#include <linux/module.h>
#include <linux/kmod.h>
#include <net/protocol.h>
#include <net/checksum.h>
#include <linux/bitops.h>
#include <linux/proc_fs.h>
#include <asm/spinlock.h>

#include "ip_nat.h"
#include "ip_nat_binding.h"
#include "ip_nat_proto.h"
#include "ip_nat_rule.h"

/* Regarding locking: we use a big spinlock for the whole lot.  We can
 * improve this by using similar techniques to modreglist.c. */
spinlock_t ip_nat_lock = SPIN_LOCK_UNLOCKED;

/* GCC rocks. */
static const char *hooknames[] = {
	[NF_IP_LOCAL_OUT] = "LOCAL-OUT",
	[NF_IP_LOCAL_IN] = "LOCAL-IN",
	[NF_IP_PRE_ROUTING] = "PRE-ROUTING",
	[NF_IP_POST_ROUTING] = "POST-ROUTING"
};

/* From ip_nat_binding.c */
static struct ip_nat_manipulation *
find_manip(const struct ip_nat_tuple_proto *tuplepr, unsigned int hook);

static int alloc_null_binding(const struct iphdr *iph,
			      const struct ip_nat_pktinfo *pkt,
			      struct ip_nat_binding **binding);

/* From ip_nat_protos.c */
static struct ip_nat_proto *find_proto(unsigned int protocol);

/* From ip_nat_rule.c */
static unsigned int ip_nat_do_rule(const struct iphdr *iph,
				   const struct ip_nat_pktinfo *pkt,
				   struct ip_nat_binding **binding);

static struct {
	/* Four hash tables for per_connection information.  Every
	 * struct ip_nat_binding is in all of them. */

	/* One rule table. */
	struct ip_nat_rule *rule_table;

	/* For tuple lookups. */
	struct ip_nat_manipulation *by_tuple[IP_NAT_HTABLE_SIZE];

	/* For finding least-allocated srcip/dstip/proto set... */
	struct ip_nat_by_ip_and_proto *by_ip_and_proto[IP_NAT_HTABLE_SIZE];

	/* For finding other mappings of this srcip/proto/per-proto-src... */
	struct ip_nat_by_source *by_source[IP_NAT_HTABLE_SIZE];

	/* For checking that a new tuple is unique. */
	struct ip_nat_by_new_tuple *by_newtuple[IP_NAT_HTABLE_SIZE];
} conns;

static struct ip_nat_manipulation **
hash_by_tuple(const struct ip_nat_tuple_proto *tuplepr)
{
	return &conns.by_tuple[csum_partial((char *)tuplepr,sizeof(*tuplepr),0)
			      % IP_NAT_HTABLE_SIZE];
}

/* Hand this the allocated tuple. */
static struct ip_nat_by_ip_and_proto **
hash_by_ip_and_proto(u_int32_t src, u_int32_t dst, u_int8_t proto)
{
	return &conns.by_ip_and_proto[(src + dst + proto)%IP_NAT_HTABLE_SIZE];
}

static struct ip_nat_by_ip_and_proto **
find_by_ip_and_proto(u_int32_t src, u_int32_t dst, u_int8_t proto,
		     struct ip_nat_by_ip_and_proto **start)
{
	struct ip_nat_by_ip_and_proto **ipp;
	
	for (ipp = start; *ipp; ipp = &(*ipp)->next)
		if ((*ipp)->srcip == src
		    && (*ipp)->dstip == dst
		    && (*ipp)->proto == proto)
			break;

	return ipp;
}

static struct ip_nat_by_source **
hash_by_source(const struct ip_nat_tuple_manip *manip, u_int32_t proto)
{
	return &conns.by_source[csum_partial((char *)manip, sizeof(*manip),
					     proto) % IP_NAT_HTABLE_SIZE];
}

/* FIXME: This hash function sucks. --RR */
static struct ip_nat_by_new_tuple **
hash_by_newtuple(const struct ip_nat_tuple_proto *tuplepr, unsigned int hook)
{
	return &conns.by_newtuple[csum_partial((char *)tuplepr,
					       sizeof(*tuplepr),
					       hook)
				 % IP_NAT_HTABLE_SIZE];
}

/* Does tuplepr + the source manip come within the range mr */
static int ip_nat_in_range(const struct ip_nat_tuple_proto *tuplepr,
			   const struct ip_nat_manip_info *manip,
			   const struct ip_nat_multi_range *mr)
{
	struct ip_nat_proto *proto = find_proto(tuplepr->proto);
	unsigned int i;
	struct ip_nat_tuple_proto new_tpr = *tuplepr;

	/* CHEAT: we know that hook == NF_IP_POST_ROUTING, so
	   manip is either none, or a single SRC manip */
	for (i = 0; i < manip->num_manips; i++) {
		if (manip->manips[i].manip_type != IP_NAT_MANIP_SRC) {
			printk("ip_nat_in_range: WTF? DST manip.\n");
			return 0;
		}
		new_tpr.tuple.src = manip->manips[i].manip;
	}

	for (i = 0; i < mr->rangesize; i++) {
		/* If we are allowed to map IPs, then we must be in the
		   range specified, otherwise we must be unchanged. */
		if (mr->range[i].flags & IP_NAT_RANGE_MAP_IPS) {
			if (ntohl(new_tpr.tuple.src.ip)
			    < ntohl(mr->range[i].min_ip)
			    || (ntohl(new_tpr.tuple.src.ip)
				> ntohl(mr->range[i].max_ip)))
				continue;
		} else {
			if (manip->manips[0].manip.ip != tuplepr->tuple.src.ip)
				continue;
		}

		if ((mr->range[i].flags & IP_NAT_RANGE_PROTO_SPECIFIED)
		    && proto->in_range(&new_tpr, IP_NAT_MANIP_SRC,
				       &mr->range[i].min, &mr->range[i].max))
				return 1;
	}
	return 0;
}

/* Only called when hook == NF_IP_POST_ROUTING, ie. SRC manip */
static struct ip_nat_by_source *
ip_nat_find_src(const struct ip_nat_tuple_proto *tuplepr,
		const struct ip_nat_multi_range *range)
{
	static struct ip_nat_by_source *i;

	for (i = *hash_by_source(&tuplepr->tuple.src, tuplepr->proto);
	     i; i = i->next) {
		if (i->proto == tuplepr->proto
		    && memcmp(&i->src, &tuplepr->tuple.src, sizeof(i->src))
		    && ip_nat_in_range(tuplepr, i->manip_info, range))
			return i;
	}
	return i;
}

static void invert_tuplepr(struct ip_nat_tuple_proto *tuplepr)
{
	u_int32_t tmp = tuplepr->tuple.src.ip;
	
	tuplepr->tuple.src.ip = tuplepr->tuple.dst.ip;
	tuplepr->tuple.dst.ip = tmp;

	find_proto(tuplepr->proto)->invert_tuple(&tuplepr->tuple);
}

/* If it's really a local destination manip, it may need to do a
   source manip too. */
static int do_extra_mangle(u_int32_t var_ip,
			   u_int32_t *other_ipp,
			   unsigned int hook)
{
	struct rtable *rt;

	/* Only do this reroute hack for packets from local processes. */
	if (hook != NF_IP_LOCAL_OUT)
		return 1;

	/* FIXME: IPTOS_TOS(iph->tos) --RR */
	if (ip_route_output(&rt, var_ip, 0, 0, 0) != 0) {
		DEBUGP("fill_in_extra_manip: Can't get route to %u.%u.%u.%u\n",
		       IP_PARTS(var_ip));
		return 0;
	}

	*other_ipp = rt->rt_src;
	ip_rt_put(rt);
	return 1;
}

/* For fragmentation handling, we want the least-used
   src-ip/dst-ip/proto triple.  Fairness doesn't come into it.  Thus
   if the range specifies 1.2.3.4 ports 10000-10005 and 1.2.3.5 ports
   1-65535, we don't do pro-rata allocation based on ports; we choose
   the ip with the lowest src-ip/dst-ip/proto usage.

   If an allocation then fails (eg. all 6 ports used in the 1.2.3.4
   range), we eliminate that and try again.  This is not the most
   efficient approach, but if you're worried about that, don't hand us
   ranges you don't really have.  */
static struct ip_nat_range *
ip_nat_find_best_ips_proto(struct ip_nat_tuple_proto *tuplepr,
			   const struct ip_nat_multi_range *mr,
			   unsigned int hook)
{
	unsigned int i;
	struct {
		const struct ip_nat_range *range;
		unsigned int score;
		struct ip_nat_tuple_proto tuplepr;
	} best = { NULL,  0xFFFFFFFF };
	u_int32_t *var_ipp, *other_ipp, saved_ip;

	if (HOOK2MANIP(hook) == IP_NAT_MANIP_SRC) {
		var_ipp = &tuplepr->tuple.src.ip;
		saved_ip = tuplepr->tuple.dst.ip;
		other_ipp = &tuplepr->tuple.dst.ip;
	} else {
		var_ipp = &tuplepr->tuple.dst.ip;
		saved_ip = tuplepr->tuple.src.ip;
		other_ipp = &tuplepr->tuple.src.ip;
	}

	for (i = 0; i < mr->rangesize; i++) {
		u_int32_t minip, maxip;

		/* Don't do ranges which are already eliminated. */
		if (mr->range[i].flags & IP_NAT_RANGE_FULL) continue;

		if (mr->range[i].flags & IP_NAT_RANGE_MAP_IPS) {
			minip = mr->range[i].min_ip;
			maxip = mr->range[i].max_ip;
		} else
			minip = maxip = *var_ipp;

		DEBUGP("ip_nat_find_best: %u.%u.%u.%u - %u.%u.%u.%u\n",
		       IP_PARTS(minip), IP_PARTS(maxip));
		for (*var_ipp = minip;
		     ntohl(*var_ipp) <= ntohl(maxip);
		     *var_ipp = htonl(ntohl(*var_ipp) + 1)) {
			struct ip_nat_by_ip_and_proto **ipp;
			unsigned int score = 0;

			/* Reset the other ip in case it was mangled by
			 * do_extra_mangle last time. */
			*other_ipp = saved_ip;

			/* Can't route?  This whole range part is
			 * probably screwed, but keep trying
			 * anyway. */
			if (!do_extra_mangle(*var_ipp, other_ipp, hook))
				continue;

			ipp = hash_by_ip_and_proto(tuplepr->tuple.src.ip,
						   tuplepr->tuple.dst.ip,
						   tuplepr->proto);
			while (*(ipp
				 = find_by_ip_and_proto(tuplepr->tuple.src.ip,
							tuplepr->tuple.dst.ip,
							tuplepr->proto,
							ipp))) {
				score++;
				ipp = &(*ipp)->next;
			}

			DEBUGP("ip_nat_find_best: score for %u.%u.%u.%u: %u\n",
			       IP_PARTS(*var_ipp), score);
			if (score < best.score) {
				DEBUGP("ip_nat_find_best_ips_proto: score %u "
				       "beats %u ip=%u.%u.%u.%u\n",
				       score, best.score,
				       IP_PARTS(*var_ipp));
				/* Optimization: doesn't get any better than
				   this. */
				if (score == 0)
					return (struct ip_nat_range *)
						&mr->range[i];

				best.score = score;
				best.tuplepr = *tuplepr;
				best.range = &mr->range[i];
			}
		}
	}
	*tuplepr = best.tuplepr;

	/* Discard const. */
	return (struct ip_nat_range *)best.range;
}

/* Fills in the tuplepr */
static int
get_unique_tuplepr(struct ip_nat_tuple_proto *tuplepr,
		   const struct ip_nat_tuple_proto *orig_tuplepr,
		   struct ip_nat_multi_range *mr,
		   unsigned int hook)
{
	struct ip_nat_proto *proto = find_proto(orig_tuplepr->proto);
	struct ip_nat_by_source *by_source;
	struct ip_nat_range *rptr;
	unsigned int i;

	/* 1) If this srcip/proto/src-proto-part is currently mapped,
	   and that same mapping gives a unique tuple within the given
	   range, use that.

	   This is only required for source (ie. NAT/masq) mappings.
	   So far, we don't do local source mappings, so multiple
	   manips not an issue.  */
	if (hook == NF_IP_POST_ROUTING
	    && (by_source = ip_nat_find_src(orig_tuplepr, mr))) {
		/* Apply that manip_info. */
		*tuplepr = ((struct ip_nat_tuple_proto)
			    { orig_tuplepr->proto,
			      { by_source->manip_info->manips[0].manip,
				orig_tuplepr->tuple.dst } });
		DEBUGP("get_unique_tuple: Found current mapping\n");
		return 1;
	}

	/* Clear full flags. */
	for (i = 0; i < mr->rangesize; i++)
		mr->range[i].flags &= ~IP_NAT_RANGE_FULL;

	/* 2) Select the least-used IP/proto combination in the given
	   range.
	*/
	*tuplepr = *orig_tuplepr;
	while ((rptr = ip_nat_find_best_ips_proto(tuplepr, mr, hook)) != NULL){
		/* 3) The per-protocol part of the manip is made to
		   map into the range to make a unique tuple. */

		/* Only bother mapping if it's not already in range
		   and unique */
		if ((!(rptr->flags & IP_NAT_RANGE_PROTO_SPECIFIED)
		     || proto->in_range(tuplepr, HOOK2MANIP(hook),
					&rptr->min, &rptr->max))
		    && !ip_nat_used_tuple(tuplepr, hook)) {
			DEBUGP("No need to remap per-proto.\n");
			return 1;
		}
		else if (proto->unique_tuple(tuplepr, rptr, hook)){
#ifdef DEBUGNAT
			if (ip_nat_used_tuple(tuplepr, hook))
				DEBUGP("Proto made nonunique tuple!\n");
#endif
			return 1;
		}

		/* Eliminate that from range, and try again. */
		rptr->flags |= IP_NAT_RANGE_FULL;
		*tuplepr = *orig_tuplepr;
	}

	return 0;
}

/* Check for tuple existence */
int ip_nat_used_tuple(const struct ip_nat_tuple_proto *tuplepr,
		      unsigned int hook)
{
	struct ip_nat_by_new_tuple **i;

	for (i = hash_by_newtuple(tuplepr, hook); *i; i = &(*i)->next)
		if (hook == (*i)->hook
		    && memcmp(tuplepr, &(*i)->tuplepr, sizeof(*tuplepr)) == 0)
			return 1;
	return 0;
}

static int get_tuple_proto(const struct iphdr *iph,
			   size_t len,
			   struct ip_nat_tuple_proto *tuplepr)
{
	IP_NAT_TUPLE_PROT_INIT(tuplepr, iph->saddr, iph->daddr, iph->protocol);

	return find_proto(tuplepr->proto)
		->pkt_to_tuple((u_int32_t *)iph + iph->ihl,
			       len - 4*iph->ihl, &tuplepr->tuple);
}

/* We do checksum mangling, so if they were wrong before they're still
 * wrong.  Also works for incomplete packets (eg. ICMP dest
 * unreachables.) */
u_int16_t ip_nat_cheat_check(u_int32_t oldvalinv, u_int32_t newval,
			     u_int16_t oldcheck)
{
	u_int32_t diffs[] = { oldvalinv, newval };
	return csum_fold(csum_partial((char *)diffs, sizeof(diffs),
				      oldcheck^0xFFFF));
}

static void ip_nat_manip_pkt(u_int16_t proto, struct iphdr *iph, size_t len,
			     const struct ip_nat_tuple_manip *manip,
			     enum ip_nat_manip_type manip_type)
{
	DEBUGP("ip_nat_manip_pkt: tuple_manip = %p.\n", manip);
	find_proto(proto)->manip_pkt(iph, len, manip, manip_type);

	if (manip_type == IP_NAT_MANIP_SRC) {
		DEBUGP("ip_nat_manip_pkt: Replacing src ip %u.%u.%u.%u with "
		       "%u.%u.%u.%u\n", IP_PARTS(iph->saddr),
		       IP_PARTS(manip->ip));

		iph->check = ip_nat_cheat_check(~iph->saddr, manip->ip,
						iph->check);
		iph->saddr = manip->ip;
	} else {
		DEBUGP("ip_nat_manip_pkt: Replacing dst ip %u.%u.%u.%u with "
		       "%u.%u.%u.%u\n", IP_PARTS(iph->daddr),
		       IP_PARTS(manip->ip));

		iph->check = ip_nat_cheat_check(~iph->daddr, manip->ip,
						iph->check);
		iph->daddr = manip->ip;
	}
}

static int icmp_reply_translation(struct sk_buff *skb,
				  unsigned int hook)
{
	struct iphdr *iph = skb->nh.iph;
	struct icmphdr *hdr = (struct icmphdr *)((u_int32_t *)iph + iph->ihl);
	struct ip_nat_tuple_proto innertuplepr;
	struct iphdr *inner = (struct iphdr *)(hdr + 1);
	size_t datalen = skb->len - iph->ihl*4 - sizeof(*hdr);
	struct ip_nat_manipulation *manip;
	unsigned int i;
	struct ip_nat_binding *binding;

	if (skb->len < iph->ihl * 4 + sizeof(struct icmphdr)) {
		DEBUGP("icmp_reply_translation: too short\n");
		return 0;
	}

	if (hdr->type != ICMP_DEST_UNREACH
	    && hdr->type != ICMP_SOURCE_QUENCH
	    && hdr->type != ICMP_TIME_EXCEEDED
	    && hdr->type != ICMP_PARAMETERPROB
	    && hdr->type != ICMP_REDIRECT)
		return 0;

	/* Ignore it if the checksum's bogus. */
	if (ip_compute_csum((unsigned char *)hdr, sizeof(*hdr) + datalen)) {
		DEBUGP("icmp_reply_translation: bad csum\n");
		return 0;
	}

	/* Are they talking about one of our connections? */
	if (!get_tuple_proto(inner, datalen, &innertuplepr)) {
		DEBUGP("icmp_reply_translation: Failed to get tuple\n");
		return 0;
	}

	/* Ordinarily, we'd expect the inverted tupleproto, but it's
	   been preserved inside the ICMP. */
	invert_tuplepr(&innertuplepr);
	manip = find_manip(&innertuplepr, hook);
	/* We can get ICMPs from the local host talking about
	   through packets. */
	if (!manip && hook == NF_IP_LOCAL_OUT)
		manip = find_manip(&innertuplepr, NF_IP_POST_ROUTING);
	if (!manip) {
		DEBUGP("icmp_reply_translation: no match\n");
		return 0;
	}

	DEBUGP("icmp_reply_translation: Found manip.\n");

	/* Map inner packet first; do this REVERSE of normal. */
	for (i = 0; i < manip->manip_info.num_manips; i++)
		ip_nat_manip_pkt(inner->protocol, inner, datalen,
				 &manip->manip_info.manips[i].manip,
				 !manip->manip_info.manips[i].manip_type);

	/* Now outer packet. */
	/* Note: May not be from a NAT'd host, but probably safest to
	   do translation always as if it came from the host itself
	   (even though a "host unreachable" coming from the host
	   itself is a bit wierd). */

	/* Find partner manip: store binding because partner may be
           expired (hence, NULL binding pointer). */
	binding = manip->binding;
	manip = &binding->manip[!HOOK2MANIP(hook)];

	/* Use that to map outer packet: 0 give no per-proto mapping */
	for (i = 0; i < manip->manip_info.num_manips; i++)
		ip_nat_manip_pkt(0, iph, skb->len,
				 &manip->manip_info.manips[i].manip,
				 manip->manip_info.manips[i].manip_type);

	/* Since we mangled inside ICMP packet, recalculate its
	   checksum from scratch.  (Hence the handling of incorrect
	   checksums above, so we don't accidentally fix one.) */
	hdr->checksum = 0;
	hdr->checksum = ip_compute_csum((unsigned char *)hdr,
					sizeof(*hdr) + datalen);
	binding->relatedpacket(binding, iph, skb->len, HOOK2MANIP(hook));
	/* If we had a hardware checksum before, it's now invalid */
	skb->ip_summed = CHECKSUM_NONE;
	return 1;
}

/* Do packet manipulations according to binding and unlock. */
static unsigned int do_binding(struct sk_buff *skb,
			       struct ip_nat_binding *binding,
			       unsigned int hook,
			       unsigned long flags)
{
	struct iphdr *iph = skb->nh.iph;
	unsigned int i;
	unsigned int ret;
	struct ip_nat_manipulation *manip;

	manip = &binding->manip[HOOK2MANIP(hook)];

	for (i = 0; i < manip->manip_info.num_manips; i++) {
		/* If we had a hardware checksum before, it's now invalid */
		skb->ip_summed = CHECKSUM_NONE;
		skb->nfcache |= NFC_ALTERED;
		DEBUGP("do_binding: %s manip number %i.\n",
		       HOOK2MANIP(hook) == IP_NAT_MANIP_SRC ? "SRC" : "DST",
		       i);
		ip_nat_manip_pkt(iph->protocol, iph, skb->len,
				 &manip->manip_info.manips[i].manip,
				 manip->manip_info.manips[i].manip_type);
	}

	manip->hitcount++;
	manip->lasthit = jiffies;
	ret = binding->postmangle(binding, iph, skb->len, HOOK2MANIP(hook));
	spin_unlock_irqrestore(&ip_nat_lock, flags);

	return ret;
}

static unsigned int ipnat(struct sk_buff *skb,
			  const struct device *device,
			  unsigned int hook)
{
	struct iphdr *iph = skb->nh.iph;
	struct ip_nat_pktinfo pkt;
	struct ip_nat_manipulation *manip;
	struct ip_nat_binding *binding = NULL;
	unsigned long flags;

	pkt.device = device;
	pkt.hook = hook;
	/* FIXME: One day, fill in properly. --RR */
	skb->nfcache |= NFC_UNKNOWN;

	spin_lock_irqsave(&ip_nat_lock, flags);
	if (!get_tuple_proto(iph, skb->len, &pkt.tuplepr)) {
		DEBUGP("ipnat: Can't get tuple.\n");
		spin_unlock_irqrestore(&ip_nat_lock, flags);
		return NF_ACCEPT;
	}

	/* ICMP replies tied to other protocols are handled separately:
	 * don't clutter binding table for such rare occurrances. */
	if (iph->protocol == IPPROTO_ICMP) {
		if (icmp_reply_translation(skb, hook)) {
			spin_unlock_irqrestore(&ip_nat_lock, flags);
			skb->nfcache |= NFC_ALTERED;
			return NF_ACCEPT;
		}
	}

	/* First, seek per-connection match. */
	manip = find_manip(&pkt.tuplepr, pkt.hook);
	if (manip) {
		binding = manip->binding;
		DEBUGP("ipnat: Found binding %p (postmangle %p).\n",
		       binding, binding->postmangle);
		return do_binding(skb, binding, hook, flags);
	}

	/* Look for rule match. */
	switch (ip_nat_do_rule(iph, &pkt, &binding)) {
	case NF_DROP:
		DEBUGP("ipnat: Rejected by rule.\n");
		spin_unlock_irqrestore(&ip_nat_lock, flags);
		return NF_DROP;

		/* FIXME: it shouldn't ever do this. */
	case NF_QUEUE:
		DEBUGP("ipnat: QUEUE by rule.\n");
		spin_unlock_irqrestore(&ip_nat_lock, flags);
		return NF_QUEUE;

	default:
		if (!binding) {
			/* Allocate a null binding if local packet. */
			if (!alloc_null_binding(iph, &pkt, &binding)) {
				DEBUGP("Can't allocate null tuple.\n");
				spin_unlock_irqrestore(&ip_nat_lock,
						       flags);
				return NF_DROP;
			}
		}
		return do_binding(skb, binding, hook, flags);
	}

}

static unsigned int ipnat_infn(unsigned int hooknum,
			       struct sk_buff **pskb,
			       const struct device *in,
			       const struct device *out)
{
	DEBUGP("ipnat_infn\n");
	return ipnat(*pskb, in, hooknum);
}

static unsigned int ipnat_outfn(unsigned int hooknum,
				struct sk_buff **pskb,
				const struct device *in,
				const struct device *out)
{
	DEBUGP("ipnat_outfn\n");
	return ipnat(*pskb, out, hooknum);
}

static unsigned int ipnat_local_infn(unsigned int hooknum,
				     struct sk_buff **pskb,
				     const struct device *in,
				     const struct device *out)
{
	DEBUGP("ipnat_local_infn\n");
	return ipnat(*pskb, NULL, hooknum);
}

static unsigned int ipnat_local_outfn(unsigned int hooknum,
				      struct sk_buff **pskb,
				      const struct device *in,
				      const struct device *out)
{
	DEBUGP("ipnat_local_outfn\n");
	return ipnat(*pskb, NULL, hooknum);
}

/* Packet filtering is priority 0: for local_in, we want to be after
   it, for local_out, we want to be before. */
static struct nf_hook_ops ipnat_in_ops = { NULL, ipnat_infn, NULL,
					   PF_INET, NF_IP_PRE_ROUTING, 0 };
static struct nf_hook_ops ipnat_out_ops = { NULL, ipnat_outfn, NULL,
					    PF_INET, NF_IP_POST_ROUTING, 0 };
static struct nf_hook_ops ipnat_local_in_ops = { NULL, ipnat_local_infn,
						 NULL,
						 PF_INET, NF_IP_LOCAL_IN,
						 100 };
static struct nf_hook_ops ipnat_local_out_ops = { NULL, ipnat_local_outfn,
						  NULL,
						  PF_INET, NF_IP_LOCAL_OUT,
						  -100 };

#include "ip_nat_proto.c"
#include "ip_nat_binding.c"
#include "ip_nat_rule.c"

#ifdef CONFIG_PROC_FS
struct proc_dir_entry ip_nat_conns_procinfo = {
	PROC_NET_NAT_BINDINGS, 15, "ip_nat_bindings", S_IFREG | S_IRUGO,
	1, 0, 0, 0, &proc_net_inode_operations, nat_get_bindings
};
struct proc_dir_entry ip_nat_rule_procinfo = {
	PROC_NET_NAT_RULES, 12, "ip_nat_rules", S_IFREG | S_IRUGO,
	1, 0, 0, 0, &proc_net_inode_operations, nat_get_rules
};
#endif

/* We need to know when devices go up and down so we can correct the
   device pointers. */
static int ipnat_device_event(struct notifier_block *this,
			      unsigned long event, void *voiddev)
{
	struct device *dev = voiddev;

	switch (event) {
	case NETDEV_CHANGENAME:
	case NETDEV_DOWN:
		ipnat_rule_down_device(dev);
		if (event == NETDEV_DOWN)
			break;

		/* Fall thru... iff CHANGENAME. */
	case NETDEV_UP:
		ipnat_rule_up_device(dev, dev->name);
		break;
	}
	return NOTIFY_DONE;
}

static struct notifier_block ipnat_dev_notifier = {
	ipnat_device_event,
	NULL,
	0
};

static int init_or_cleanup(int init)
{
	int ret = 0;
	unsigned long flags;

	spin_lock_irqsave(&ip_nat_lock, flags);
	
	if (!init) goto cleanup;

#ifdef CONFIG_PROC_FS
	proc_net_register(&ip_nat_rule_procinfo);
	proc_net_register(&ip_nat_conns_procinfo);
#endif
	ret = nf_register_hook(&ipnat_in_ops);
	if (ret < 0) {
		printk("ip_nat_init: can't register in hook.\n");
		goto cleanup_nothing;
	}
	ret = nf_register_hook(&ipnat_out_ops);
	if (ret < 0) {
		printk("ip_nat_init: can't register out hook.\n");
		goto cleanup_inops;
	}
	ret = nf_register_hook(&ipnat_local_in_ops);
	if (ret < 0) {
		printk("ip_nat_init: can't register local in hook.\n");
		goto cleanup_outops;
	}
	ret = nf_register_hook(&ipnat_local_out_ops);
	if (ret < 0) {
		printk("ip_nat_init: can't register local out hook.\n");
		goto cleanup_localinops;
	}
	ret = init_ip_nat_rule();
	if (ret < 0) {
		printk("ip_nat_init: can't setup rules.\n");
		goto cleanup_localoutops;
	}
	init_protos();	
	register_netdevice_notifier(&ipnat_dev_notifier);

	spin_unlock_irqrestore(&ip_nat_lock, flags);
	return ret;

 cleanup:
	/* Cleanup protocols first, otherwise we dereg proto rules twice. */
	cleanup_protos();
	cleanup_ip_nat_rule();
	cleanup_ip_nat_null_bindings();
	unregister_netdevice_notifier(&ipnat_dev_notifier);
 cleanup_localoutops:
	nf_unregister_hook(&ipnat_local_out_ops);
 cleanup_localinops:
	nf_unregister_hook(&ipnat_local_in_ops);
 cleanup_outops:
	nf_unregister_hook(&ipnat_out_ops);
 cleanup_inops:
	nf_unregister_hook(&ipnat_in_ops);
 cleanup_nothing:
	spin_unlock_irqrestore(&ip_nat_lock, flags);
	return ret;
}

#ifdef MODULE
int init_module(void)
{
	return init_or_cleanup(1);
}

void cleanup_module(void)
{
#ifdef CONFIG_PROC_FS
	proc_net_unregister(PROC_NET_NAT_RULES);
	proc_net_unregister(PROC_NET_NAT_BINDINGS);
#endif
	init_or_cleanup(0);
}
#else
__initfunc(void ip_nat_init(void))
{
	if (init_or_cleanup(1) < 0)
		panic("Can't set up ip_nat.\n");
}
#endif /*MODULE*/
