#ifndef _IP_NAT_H
#define _IP_NAT_H

struct sk_buff;
struct device;
struct iphdr;

#ifdef DEBUGNAT
#define DEBUGP printk
#else
#define DEBUGP(format, args...)
#endif

/* The protocol-specific parts of the tuple which we can manupulate. */
union ip_nat_proto_manip
{
	struct {
		u_int16_t port;
	} tcp;
	struct {
		u_int16_t port;
	} udp;
	struct {
		u_int16_t id;
	} icmp;
	/* Add other protocols here. */

	u_int32_t all[0];
};

/* These are the parts of the tuple we can manipulate (for
 * outgoing tuple allocation) */
struct ip_nat_tuple_manip
{
	u_int32_t ip;
	union ip_nat_proto_manip u;
};

/* This contains the information to distinguish a connection. */
struct ip_nat_tuple
{
	struct ip_nat_tuple_manip src;

	/* These are the parts of the tuple which are fixed for
	 * outgoing tuple allocation. */
	struct {
		u_int32_t ip;
		union {
			struct {
				u_int16_t port;
			} tcp;
			struct {
				u_int16_t port;
			} udp;
			struct {
				u_int8_t type, code;
			} icmp;
			/* Add other protocols here. */

			u_int32_t all[0];
		} u;
	} dst;
};

struct ip_nat_tuple_proto
{
	u_int32_t proto;
	struct ip_nat_tuple tuple;
};

struct ip_nat_pktinfo
{
	const struct device *device;
	unsigned int hook;
	struct ip_nat_tuple_proto tuplepr;
};

#define HOOK2MANIP(hook)					\
((hook) == NF_IP_PRE_ROUTING || (hook) == NF_IP_LOCAL_OUT	\
 ? IP_NAT_MANIP_DST : IP_NAT_MANIP_SRC)

enum ip_nat_direction
{
	IP_NAT_INCOMING,
	IP_NAT_OUTGOING
};

enum ip_nat_manip_type
{
	IP_NAT_MANIP_SRC = IP_NAT_OUTGOING,
	IP_NAT_MANIP_DST = IP_NAT_INCOMING
};

#define IP_NAT_RANGE_MAP_IPS 1
#define IP_NAT_RANGE_PROTO_SPECIFIED 2
/* Used internally by get_unique_tuple(). */
#define IP_NAT_RANGE_FULL 4

/* Single range specification. */
struct ip_nat_range
{
	/* Set to OR of flags above. */
	unsigned int flags;

	/* Inclusive: network order. */
	u_int32_t min_ip, max_ip;

	union ip_nat_proto_manip min, max;
};

/* A range consists of an array of ip_nat_range */
struct ip_nat_multi_range
{
	unsigned int rangesize;

	struct ip_nat_range range[0];
};

/* Calculates relative checksums. */
extern u_int16_t ip_nat_cheat_check(u_int32_t oldvalinv, u_int32_t newval,
				    u_int16_t oldcheck);

/* Check for tuple existence. */
extern int ip_nat_used_tuple(const struct ip_nat_tuple_proto *tuplepr,
			     unsigned int hook);

#ifdef __KERNEL__
#define IP_PARTS_NATIVE(n)			\
(unsigned int)((n)>>24)&0xFF,			\
(unsigned int)((n)>>16)&0xFF,			\
(unsigned int)((n)>>8)&0xFF,			\
(unsigned int)((n)&0xFF)

#define IP_PARTS(n) IP_PARTS_NATIVE(ntohl(n))

#define IP_NAT_TUPLE_PROT_INIT(t,_src,_dst,pr)		\
do {							\
	memset((t), 0, sizeof(*(t)));			\
	(t)->proto = (pr);				\
	(t)->tuple.src.ip = (_src);			\
	(t)->tuple.dst.ip = (_dst);			\
} while(0)

static inline void dump_tuplepr(const char *fn,
				const struct ip_nat_tuple_proto *tuplepr,
				unsigned int hook)
{
#ifdef DEBUGNAT
	printk("%s: %s proto=%u src=%u.%u.%u.%u "
	       "dst=%u.%u.%u.%u ",
	       fn,
	       hook == NF_IP_PRE_ROUTING ? "PRE-ROUTING"
	       : (hook == NF_IP_POST_ROUTING ? "POST-ROUTING"
		  : (hook == NF_IP_LOCAL_OUT ? "LOCAL-OUT"
		     : (hook == NF_IP_LOCAL_IN ? "LOCAL-IN" : "ANY"))),
	       tuplepr->proto, IP_PARTS(tuplepr->tuple.src.ip),
	       IP_PARTS(tuplepr->tuple.dst.ip));

	switch (tuplepr->proto) {
	case IPPROTO_ICMP:
		printk("id=%u type=%u code=%u\n",
		       ntohs(tuplepr->tuple.src.u.icmp.id),
		       tuplepr->tuple.dst.u.icmp.type,
		       tuplepr->tuple.dst.u.icmp.code);
		break;
	case IPPROTO_UDP:
	case IPPROTO_TCP:
		printk("sport=%u dport=%u\n",
		       ntohs(tuplepr->tuple.src.u.udp.port),
		       ntohs(tuplepr->tuple.dst.u.udp.port));
		break;
	}
#endif
}

static inline void dump_iph(const char *fn, const struct iphdr *iph)
{
#ifdef DEBUGNAT
	printk("%s: proto=%u src=%u.%u.%u.%u dst=%u.%u.%u.%u\n",
	       fn, iph->protocol, IP_PARTS(iph->saddr), IP_PARTS(iph->daddr));
#endif
}

static inline void dump_manip(const char *fn,
			      enum ip_nat_manip_type mtype,
			      const struct ip_nat_tuple_manip *manip)
{
#ifdef DEBUGNAT
	printk("%s manip: %s=%u.%u.%u.%u id=%u\n",
	       fn, mtype == IP_NAT_MANIP_SRC ? "src" : "dst",
	       IP_PARTS(manip->ip), ntohs(manip->u.icmp.id));
#endif
}
#endif
#endif /*_IP_NAT_H*/
